from _dittoffi import lib, ffi
from dittolive.exceptions import FfiException

class WriteTransaction:
    """Write transaction class.

    Use it with the `with` keyword to have a scoped transaction.
    """
    def __init__(self, raw_ditto):
        self.__raw_ditto = raw_ditto

        res = lib.ditto_write_transaction(self.__raw_ditto, ffi.NULL)
        if res.status_code != 0:
            raise FfiException

        self.__txn = res.txn

    def __enter__(self):
        return self.__txn

    def __exit__(self, exc_type, exc_value, exc_traceback):
        status = lib.ditto_write_transaction_commit(self.__raw_ditto, self.__txn)
        if status != 0:
            raise FfiException



"""Module identity."""

from abc import ABC, abstractmethod
from _dittoffi import lib

class Identity(ABC):
    """Abstract class for all identities."""

    def __init__(self, app_id:str):
        """Init motherclass of all Identities."""
        self.__app_id = app_id
        self._auth_client = None

    def app_id(self):
        """Returns the AppId currently in use by this peer."""
        return self.__app_id

    def site_id(self):
        """Returns the current SiteId identifying the Ditto peer."""
        raise NotImplementedError

    def auth_client(self):
        """Returns a shared reference to the underlying AuthClient."""
        if self._auth_client is None :
            raise ValueError("No `auth_client` available")
        return self._auth_client

    @abstractmethod
    def _authenticate(self, working_dir: str):
        """Auth a given type of Identity."""

    def is_web_valid(self) -> bool:
        """Returns if the current web auth token is valid."""
        return lib.ditto_auth_client_is_web_valid(self.auth_client()) == 1

    def is_x509_valid(self) -> bool:
        """Returns if the configured x509 certificate is valid."""
        return lib.ditto_auth_client_is_x509_valid(self.auth_client()) == 1

    def close(self):
        """Close the identity."""
        if self._auth_client is None:
            return

        lib.ditto_auth_client_free(self._auth_client)
        self._auth_client = None

    def requires_offline_license_token(self):
        """Return true if the Identity requires an offline license token."""
        return False

    def is_cloud_sync_enabled(self):
        """Return true is the connection with Ditto cloud is enabled."""
        return False

    def _is_online(self):
        """Return true is the Identity should have an access to Ditto cloud."""
        return False

    def _sync_url(self) -> str:
        return f"wss://{self.app_id()}.cloud.ditto.live"

    def _auth_url(self) -> str:
        return f"https://{self.app_id()}.cloud.ditto.live"



from dittolive import *
from dittolive import TransportConfig
from dittolive.identity import OfflinePlayground, OnlinePlayground

import tempfile
import uuid

APP_ID = "1eea99ab-0d02-4f6f-8b33-c8c8fff74d60"

PLAYGROUND_TOKEN = "e7b5606a-4574-43f3-8caa-2baff0889837"

TEST_LICENSE: str = "\
    o2d1c2VyX2lkc2ludGVybmFsQGRpdHRvLmxpdmVmZXhwaXJ5dDIwMjUtMTItMzFUMDA\
    6MDA6MDBaaXNpZ25hdHVyZXhYSnZlcDBWT1JwSXdHY2l5REZKaVBUUEJaTVlaN0lZR\
    XhGUFNVc2I3OTRwcjhwVjdlNjRQUkRJQ1NFNjRWeGozZFc0em5QU3Z2THBxMHBkZmh\
    kOFc4YlE9PQ==\
"

def create_offline_ditto(*, working_dir :str = None, identity = OfflinePlayground("AppId", 42), transport_config: TransportConfig = TransportConfig()):
    """"
    Create an offline ditto instance
    """

    if not working_dir:
        tmp_path = tempfile.gettempdir()
        random_folder = str(uuid.uuid4())
        working_dir = f"{tmp_path}/{random_folder}/ditto"

    ditto = Ditto(identity, working_dir = working_dir, transport_config = transport_config)
    ditto.set_offline_only_license_token(TEST_LICENSE)

    return ditto

def create_online_ditto(*, working_dir :str = None, 
                        app_id: str = APP_ID,
                        token:str = PLAYGROUND_TOKEN,
                        transport_config: TransportConfig = TransportConfig()):
    """"
    Create an online playground ditto instance
    """

    identity = OnlinePlayground(app_id, token, cloud_sync_enabled = True)

    if not working_dir:
        tmp_path = tempfile.gettempdir()
        random_folder = str(uuid.uuid4())
        working_dir = f"{tmp_path}/{random_folder}/ditto"

    ditto = Ditto(identity, working_dir = working_dir, transport_config = transport_config)
    ditto.set_offline_only_license_token(TEST_LICENSE)

    return ditto



"""OfflinePlayground is an Identity target for development."""
from _dittoffi import lib

from .identity import Identity
from dittolive._utils import char_p
from dittolive.exceptions import FfiException

class OfflinePlayground(Identity):
    """OfflinePlayground is an Identity target for development."""
    def __init__(self, app_id: str, site_id: int):
        """Init an Offline playground."""
        super().__init__(app_id)
        self.__site_id = site_id

    def _authenticate(self, working_dir:str):
        """Authenticate the offline playground.

        Args:
            working_dir (str): Working dir path

        Raises:
            FfiException: Raised if auth fails
        """
        working_dir_bytes = char_p(working_dir)
        app_id_bytes = char_p(self.app_id())
        result = lib.ditto_auth_client_make_for_development(
            working_dir_bytes,
            app_id_bytes,
            self.__site_id)

        if result.status_code != 0:
            raise FfiException

        self._auth_client = result.auth_client

    def _requires_offline_license_token(self) -> bool:
        """Return if the offline license is required.

        Returns:
            bool : True if license is required
        """
        return True



"""Path of a MutableDocument."""
from _dittoffi import lib
from .abstract_document_path import AbstractDocumentPath
from dittolive._utils import char_p, cbor_slice_ref_uint8
from dittolive.exceptions import FfiException
from .update_result import UpdateResultSet, UpdateResultRemoved

class MutableDocumentPath(AbstractDocumentPath):
    """Mutable Document path."""

    def __init__(self, doc_ptr, doc_id, key, results):
        """init."""
        super().__init__(doc_ptr, doc_id, key)
        self.results = results

    """Path of a MutableDocument."""
    def get_attachment_token(self):
        """Get attachment token.

        Returns the value at the previously specified key in the document as
        an `AttachmentToken`. If the key was invalid the return is `None`.
        """
        raise NotImplementedError

    def get_counter(self):
        """Get mutable counter.

        Returns the value at the previously specified key in the document as
        a `DittoCounter`. If the key is invalid or the value is not a `DittoCounter`
        the return is `None`.
        """
        raise NotImplementedError

    def get_register(self):
        """Get mutable register.

        Returns the value at the previously specified key in the document as
        a `DittoRegister`. If the key is invalid or the value is not a `DittoRegister`
        the return is `None`.
        """
        raise NotImplementedError

    @AbstractDocumentPath.value.setter
    def value(self, value):
        """Set a value to the current path.

        Args:
            value (_type_): Should be cbor serializable
        """
        self.set(value)

    def set(self, value, is_default:bool = False):
        """Set value at current path.

        Set a value at the document's key defined by the preceding
        subscripting.
        """
        cbor_slice = cbor_slice_ref_uint8(value)
        path = char_p(self._path)

        if is_default:
            res = lib.ditto_document_set_cbor_with_timestamp(self._doc_ptr,
                    path,
                    cbor_slice[0],
                    0)
        else:
            res = lib.ditto_document_set_cbor(self._doc_ptr, path, cbor_slice[0])

        if res != 0:
            raise FfiException

        self.results.append(UpdateResultSet(self._doc_id, self._path, value))

    def remove(self):
        """Remove value at current path.

        Remove a value at the document's key defined by the preceding
        subscripting.
        """
        lib.ditto_document_remove(self._doc_ptr, char_p(self._path))
        self.results.append(UpdateResultRemoved(self._doc_id, self._path))

    def __setitem__(self, key, value):
        """Set value.

        Args:
            key (_type_): path to the field
            value (_type_): value to be set
        """
        self[key].value = value



"""All exceptions of the dittolive package."""
from _dittoffi import lib, ffi

from dittolive._utils import string
class DittoException(Exception):
    """Superclass of all dittolive exceptions.

    It will never be raised directly.
    """

class FfiException(DittoException):
    """Something went wrong on the other side of the FFI."""
    def __init__(self):
        """Create a new FfiException."""
        message = lib.ditto_error_message()
        if message == ffi.NULL:
            super().__init__("Error across the FFI")
        else :
            str_message = string(message)
            super().__init__(f"Error across the FFI : {str_message}")

NOT_FOUND_ERROR_CODE = -30798
class DocumentNotFoundException(DittoException):
    """Document was not found."""

class ConfigException(DittoException):
    """Bad config."""

class LicenseException(DittoException):
    """License was not validated for some reason."""

class NotAuthentifiedException(DittoException):
    """The current instance is not authentified."""

class NotActivatedException(DittoException):
    """Ditto requires a valid license."""



"""Mutable documents."""

import cbor2

from _dittoffi import lib
from dittolive._utils import bytes_from_boxed

from .document_id import DocumentId
from .mutable_document_path import MutableDocumentPath

class MutableDocument:
    """TODO."""
    def __init__(self, doc_ptr, doc_id: DocumentId):
        """Create a new mutable document."""
        self.__doc_ptr = doc_ptr
        self.__doc_id = doc_id
        self.results = []

    def __getitem__(self, key:str):
        """Get mutable path in the document.

        Args:
            key (str): _description_.
        """
        return MutableDocumentPath(self.__doc_ptr, self.__doc_ptr, key, self.results)

    def __setitem__(self, key, value):
        """Set value.

        Args:
            key (_type_): path to the field
            value (_type_): value to be set
        """
        self[key].value = value

    @property
    def value(self) -> dict:
        """Get the document's inner value."""
        cbor = lib.ditto_document_cbor(self.__doc_ptr)
        cbor_bytes = bytes_from_boxed(cbor)

        return cbor2.loads(cbor_bytes)



from _dittoffi import lib
from .transport_config import TransportConfig, LanConfig, TcpConfig
from .sync_state import SyncState
from dittolive.identity import Identity
from dittolive._utils import char_p

class TransportSync:
    """Transport sync."""
    def __init__(self, config : TransportConfig, ditto, identity: Identity):

        self.__requested_state = SyncState(config, identity)
        self.__effective_state = SyncState(TransportConfig(), identity)

        self.__ws_clients = dict()

        self.mdns_client_transporter = None
        self.mdns_server_advertiser = None

        self.ble_client_transport = None
        self.ble_server_transport = None

        self.__raw_ditto = ditto

    def _update(self):
        """Update Transport layer using provided parameter.

        Afterward, it also update the effective configuration
        and applies changes.
        """
        future_state = self.__requested_state.compute_effective_state()
        self.apply_transport_state(future_state, self.__effective_state)
        self.apply_transport_global_state(future_state, self.__effective_state)
        self.__effective_state = future_state

    def apply_transport_state(self, future:SyncState, old: SyncState):
        self.__update_peer_to_peer_bluetooth_le(future, old)
        self.__update_peer_to_peer_lan(future, old)
        self.__update_listen_tcp(future, old)
        self.__update_listen_http(future, old)
        self.__update_connect_tcp_servers(future, old)
        self.__update_connect_websocket_url(future, old)
        self.__update_connect_retry_interval(future, old)

    def start_sync(self):
        """Start sync in the config."""
        self.__requested_state.sync = True
        self._update()

    def stop_sync(self):
        """Stop sync."""
        self.__requested_state.sync = False
        self._update()

    @property
    def transport_config(self):
        return self.__requested_state.transport_config

    def apply_transport_global_state(self, future: SyncState, old:SyncState):
        new_sync_group = future.transport_config.sync_group
        old_sync_group = old.transport_config.sync_group

        if new_sync_group != old_sync_group:
            lib.ditto_set_sync_group(self.__raw_ditto, new_sync_group)

    def __update_peer_to_peer_lan(self, future:SyncState, old:SyncState):
        """Apply changes in the lan config."""

    def __update_peer_to_peer_bluetooth_le(self, future:SyncState, old:SyncState):
        """Apply changes in the BLE config."""

    def __update_listen_tcp(self, future:SyncState, old:SyncState):
        """Apply changes in the TCP config."""

    def __update_listen_http(self, future:SyncState, old:SyncState):
        """Apply changes in the HTTP config."""

    def __update_connect_tcp_servers(self, future:SyncState, old:SyncState):
        """Apply changes in the TCP servers config."""

    def __update_connect_websocket_url(self, future:SyncState, old:SyncState):
        """Apply changes in the websocket url."""
        old_addresses = old.transport_config.connect.websocket_urls
        new_addresses = future.transport_config.connect.websocket_urls
        for address in old_addresses.difference(new_addresses):
            self.__stop_ws(address)

        routing_hint = self.transport_config._global_config.routing_hint
        for address in new_addresses.difference(old_addresses):
            self.__start_ws(address, routing_hint)

    def __stop_ws(self, address):
        handle = self.__ws_clients.pop(address)
        lib.websocket_client_free_handle(handle)

    def __start_ws(self, address:str, routing_hint:int):
        handle = lib.ditto_add_websocket_client(self.__raw_ditto,
                                                char_p(address), routing_hint)
        self.__ws_clients[address] = handle

    def __update_connect_retry_interval(self, future:SyncState, old:SyncState):
        """Apply changes in connect retry interval."""



import pytest

from dittolive import *
from utils import create_offline_ditto
from dittolive.store.update_result import UpdateResultSet, UpdateResultRemoved

def test_document_id():
    ditto = create_offline_ditto()
    collection = ditto.store["test"]
    doc_id = collection.upsert({"key": "value"})

    doc = collection.find_by_id(doc_id).exec()

    assert doc.id() == doc_id

def test_document_value():
    ditto = create_offline_ditto()
    collection = ditto.store["test"]

    doc_id = b"documentId"
    content = {"_id": doc_id, "key" : "value", "other": 3, "array": [1, "two", '𒊺']}
    doc_id = collection.upsert(content)

    doc = collection.find_by_id(doc_id).exec()
    assert doc.value == content

def test_document_path():
    ditto = create_offline_ditto()
    collection = ditto.store["test"]
    content = {"key" : "value"}
    doc_id = collection.upsert(content)
    doc = collection.find_by_id(doc_id).exec()
    assert "value" == doc["key"].value

def test_composed_document_path():
    ditto = create_offline_ditto()
    collection = ditto.store["test"]
    content = {"key" : {"sub_key": "value"}}
    doc_id = collection.upsert(content)
    doc = collection.find_by_id(doc_id).exec()
    assert "value" == doc["key"]["sub_key"].value

def test_update_new_key():
    ditto = create_offline_ditto()
    collection = ditto.store["test"]
    content = {"key" : {"sub_key": "value"}}
    doc_id = collection.upsert(content)

def test_remove_path_from_document():
    ditto = create_offline_ditto()
    collection = ditto.store["test"]
    content = {"key" : {"sub_key": "value"}}
    doc_id = collection.upsert(content)
    updater = lambda mut_doc: mut_doc["key"].remove()
    collection.find_by_id(doc_id).update(updater)

    doc = collection.find_by_id(doc_id).exec()
    assert doc["key"].value is None

def test_remove_non_existing_path_from_document():
    ditto = create_offline_ditto()
    collection = ditto.store["test"]
    content = {"key" : "value"}
    doc_id = collection.upsert(content)

    updater = lambda mut_doc: mut_doc["potatoes"].remove()
    collection.find_by_id(doc_id).update(updater)

    doc = collection.find_by_id(doc_id).exec()
    assert doc["key"].value == "value"
    assert doc["potatoes"].value is None

def test_update_result():
    ditto = create_offline_ditto()
    collection = ditto.store["test"]
    content = {"key" : "value"}
    doc_id = collection.upsert(content)

    def updater(mut_doc):
        assert len(mut_doc.results) == 0
        mut_doc["key"] = 3
        mut_doc["other_key"] = "perlinpinpin"
        mut_doc["key"].remove()
        assert len(mut_doc.results) == 3
        assert isinstance(mut_doc.results[0], UpdateResultSet)
        assert mut_doc.results[0].value == 3
        assert mut_doc.results[0].path == "key"
        assert isinstance(mut_doc.results[1], UpdateResultSet)
        assert mut_doc.results[1].value == "perlinpinpin"
        assert mut_doc.results[1].path == "other_key"
        assert isinstance(mut_doc.results[1], UpdateResultSet)
        assert isinstance(mut_doc.results[2], UpdateResultRemoved)

    collection.find_by_id(doc_id).update(updater)

    doc = collection.find_by_id(doc_id).exec()
    assert doc["key"].value is None
    assert doc["other_key"].value == "perlinpinpin"


from dittolive import *
from utils import create_offline_ditto

def test_find_by_id_document():
    ditto = create_offline_ditto()
    collection = ditto.store["test"]
    doc_id = collection.upsert({"key": "value"})

    doc = collection.find_by_id(doc_id).exec()
    assert doc != None


"""Abstract Document Path.

This centralize the duplicated logic between DocumentPath and MutableDocumentPath.
"""

from abc import ABC, abstractmethod
import cbor2
from typing import Union

from _dittoffi import lib
from dittolive.exceptions import FfiException
from dittolive._utils import char_p, bytes_from_boxed
from .document_id import DocumentId

class AbstractDocumentPath(ABC):
    """Path in a document."""
    def __init__(self, doc_ptr, doc_id: DocumentId, key):
        """Init a document path."""
        self._path = key
        self._doc_id = doc_id
        self._doc_ptr = doc_ptr

    # use key: str | int when upgrading to Python 3.10
    def get(self, key: Union[int, str]):
        """Get subpath.

        Use to specify a path to a key in the document that you can
        subscript further to access a nested key in the document.
        """
        if isinstance(key, int):
            self._path += f"[{key}]"
        else:
            if not key:
                return self
            self._path += f".{key}"

        return self

    # use key: str | int when upgrading to Python 3.10
    def __getitem__(self, key):
        """Get subpath.

        Use to specify a path to a key in the document that you can
        subscript further to access a nested key in the document.
        """
        return self.get(key)

    def __value_with_type(self, path_type: int) -> dict:
        cbor_res = lib.ditto_document_get_cbor_with_path_type(self._doc_ptr,
                                                char_p(self._path),
                                                path_type)

        if cbor_res.status_code != 0:
            raise FfiException

        cbor_bytes = bytes_from_boxed(cbor_res.cbor)
        if cbor_bytes is None:
            return None

        return cbor2.loads(cbor_bytes)

    @property
    def value(self) -> dict:
        """Return value at given path.

        Returns:
            object: dict representation of your object.
            None: if the path doesn't exist

        Raises:
            FfiException
        """
        return self.__value_with_type(lib.PATH_ACCESSOR_TYPE_ANY)

    @abstractmethod
    def get_attachment_token(self) -> dict:
        """Get inner attachment token.

        Returns the value at the previously specified key in the document as
        an `AttachmentToken`. If the key was invalid the return is `None`.
        """
        return self.__value_with_type(lib.PATH_ACCESSOR_TYPE_ATTACHMENT)

    @abstractmethod
    def get_counter(self) -> dict :
        """Get inner counter.

        Returns the value at the previously specified key in the document as
        a `DittoCounter`. If the key is invalid or the value is not a `DittoCounter`
        the return is `None`.
        """
        return self.__value_with_type(lib.PATH_ACCESSOR_TYPE_COUNTER)

    @abstractmethod
    def get_register(self) -> dict :
        """Get inner register.

        Returns the value at the previously specified key in the document as
        a `DittoRegister`. If the key is invalid or the value is not a `DittoRegister`
        the return is `None`.
        """
        return self.__value_with_type(lib.PATH_ACCESSOR_TYPE_REGISTER)



"""Path in a document."""

from .abstract_document_path import AbstractDocumentPath
class DocumentPath(AbstractDocumentPath):
    """Read only view of a Document at specified path."""
    def get_attachment_token(self):
        """Get inner attachment token.

        Returns the value at the previously specified key in the document as
        an `AttachmentToken`. If the key was invalid the return is `None`.
        """
        raise NotImplementedError

    def get_counter(self):
        """Get inner counter.

        Returns the value at the previously specified key in the document as
        a `DittoCounter`. If the key is invalid or the value is not a `DittoCounter`
        the return is `None`.
        """
        raise NotImplementedError

    def get_register(self):
        """Get inner register.

        Returns the value at the previously specified key in the document as
        a `DittoRegister`. If the key is invalid or the value is not a `DittoRegister`
        the return is `None`.
        """
        raise NotImplementedError



"""Types alias for the store module."""

from enum import Enum

from _dittoffi import lib

_FIND_ALL_QUERY = "true"

class WriteStrategy(Enum):
    """Write Strategy enum."""
    Merge = lib.WRITE_STRATEGY_RS_MERGE
    InsertIfAbsent = lib.WRITE_STRATEGY_RS_INSERT_IF_ABSENT
    InsertDefaultIfAbsent = lib.WRITE_STRATEGY_RS_INSERT_DEFAULT_IF_ABSENT



"""Collection module."""

from typing import Callable
from _dittoffi import lib, ffi

from .types import WriteStrategy
from .document_id import DocumentId
from .pending_id_specific_operation import PendingIdSpecificOperation
from dittolive._utils import (cbor_slice_ref_uint8, empty_slice_ref_uint8,
    bytes_from_boxed)
from dittolive.exceptions import FfiException

class Collection:
    """Manage documents from here."""

    def __init__(self, ditto, char_p_collection_name:bytes):
        """Create a new Collection."""
        self.__raw_ditto = ditto
        self.__collection_name = char_p_collection_name

    def name(self) -> str:
        """Return a str reprensentation of the collection name."""
        str.encode(self.__collection_name)

    def find_all(self):
        """Find all collections in the store.

        Generates a DittoPendingCursorOperation that can be used to find all
        documents in the collection at a point in time or you can chain a call
        to observe_local or subscribe if you want to get updates about documents
        in the collection over time. It can also be used to update, remove, or
        evict documents.
        """
        raise NotImplementedError

    def find(self, query:str):
        """Find documents that match the given query.

        Generates a DittoPendingCursorOperation with the provided query that can
        be used to find the documents matching the query at a point in time or
        you can chain a call to observe_local or subscribe if you want to get
        updates about documents matching the query as they occur. It can also be
        used to update, remove, or evict documents.
        """
        raise NotImplementedError

    def find_with_args(self, query: str, args):
        """Find documents that match the given query and args.

        Generates a DittoPendingCursorOperation with the provided query and
        query arguments that can be used to find the documents matching the
        query at a point in time or you can chain a call to observe_local or
        subscribe if you want to get updates about documents matching the query
        as they occur. It can also be used to update, remove, or evict
        documents.

        This is the recommended function to use when performing queries on a
        collection if you have any dynamic data included in the query string. It
        allows you to provide a query string with placeholders, in the form of
        `$args.my_arg_name`, along with an accompanying dictionary of arguments,
        in the form of `{ "my_arg_name": "some value" }`, and the placeholders
        will be appropriately replaced by the matching provided arguments from
        the dictionary. This includes handling things like wrapping strings in
        quotation marks and arrays in square brackets, for example.
        """
        raise NotImplementedError

    def find_by_id(self, doc_id: DocumentId):
        """Find the document with the given doc_id.

        Generates a DittoPendingIDSpecificOperation with the provided document
        ID that can be used to find the document at a point in time or you can
        chain a call to observe_local or subscribe if you want to get updates
        about the document over time. It can also be used to update, remove, or
        evict the document.
        """
        return PendingIdSpecificOperation(self.__raw_ditto, self.__collection_name,
                                        doc_id)

    def upsert(self, content, *,
            write_strategy: WriteStrategy = WriteStrategy.Merge)-> DocumentId:
        """Inserts a new document into the collection and returns its ID.

        If the document already exists, the provided document content will be merged
        with the existing document's content.
        """
        data = cbor_slice_ref_uint8(content)

        id_res = lib.ditto_collection_insert_value(
            self.__raw_ditto,
            self.__collection_name,
            data[0],
            empty_slice_ref_uint8()[0], #doc id
            write_strategy.value, #Write Strategy Merge TODO
            ffi.NULL, #hint
            ffi.NULL, #Write txn
        )

        if id_res.status_code != 0:
            raise FfiException

        return DocumentId(bytes_from_boxed(id_res.id))

    def new_attachment(self, path:str, metadata:dict):
        """Creates a new attachment, which can then be inserted into a document.

        The file residing at the provided path will be copied into the
        Ditto’s store. The DittoAttachment object that is returned is what you
        can then use to insert an attachment into a document.

        You can provide metadata about the attachment, which will be replicated
        to other peers alongside the file attachment.

        Below is a snippet to show how you can use the new_attachment
        functionality to insert an attachment into a document.
        """
        raise NotImplementedError

    def fetch_attachment(self, token, on_change: Callable[[str],None]):
        """Fetch the attachment corresponding to the provided attachment token.

        Args:
            token (_type_): Attachment token
            on_change (Callable[[str],None]): A closure that will be called when the
                status of the request to fetch the attachment has changed.
                If the attachment is already available then this will be called almost
                immediatly with a completed status value.
        """
        raise NotImplementedError




DEFAULT_TCP_IPV6 = "[::]"
NO_PREFERRED_ROUTE_HINT = 0

class TransportConfig:
    def __init__(self):
        self.peer_to_peer = PeerToPeer()
        self.connect = Connect()
        self.listen = Listen()
        self._global_config = Global()

    @property
    def bluetooth_enabled(self) -> bool:
        """Return True if bluetooth is enabled.

        Returns:
            bool: bluetooth is enabled
        """
        return self.peer_to_peer.ble.enabled

    @bluetooth_enabled.setter
    def bluetooth_enabled(self, enabled:bool):
        """Set if bluetooth is enabled or not.

        Args:
            enabled (bool): enable Bluetooth or not
        """
        self.peer_to_peer.ble.enabled = enabled

    @property
    def sync_group(self) -> int:
        """Return sync group number.

        Returns:
            int: SyncGroup
        """
        return self._global_config.sync_group

    @sync_group.setter
    def sync_group(self, sync_group:int):
        """Set sync group.

        Args:
            sync_group (int): Sync group
        """
        self._global_config.sync_group = sync_group

class PeerToPeer:
    def __init__(self):
        self.ble = BluetoothLEConfig()
        self.lan = LanConfig()

class BluetoothLEConfig:
    def __init__(self):
        self.enabled = False

class LanConfig:
    def __init__(self):
        self.enabled = False
        self.mdns_enabled = True
        self.multicast_enabled = True

class Connect:
    def __init__(self):
        self.tcp_servers = set()
        self.websocket_urls = set()
        # duration in seconds
        self.retry_interval = 5

class Listen:
    def __init__(self):
        self.tcp = TcpConfig()
        self.http = HttpListenConfig()

class TcpConfig:
    def __init__(self):
        self.enabled = False
        self.interface_ip = DEFAULT_TCP_IPV6
        self.port = 4040

class HttpListenConfig:
    def __init__(self):
        self.enabled = False
        self.interface_ip = DEFAULT_TCP_IPV6
        self.port = 80
        self.static_content_path = None
        self.websocket_sync = True
        self.tls_key_path = None
        self.tls_certificate_path = None

class Global:
    def __init__(self):
        self.sync_group = 0
        self.routing_hint = NO_PREFERRED_ROUTE_HINT



"""State of TransportConfig."""

from .transport_config import TransportConfig
from dittolive.identity import Identity
import copy
class SyncState:
    def __init__(self, config: TransportConfig, identity: Identity):
        self.__sync_active = False
        self.__web_valid = identity.is_web_valid()
        self.__x509_valid = identity.is_x509_valid()
        self.__identity = identity
        self.transport_config = config

    @property
    def sync(self):
        self.__sync_active()

    @sync.setter
    def sync(self, activated: bool):
        self.__sync_active = activated

    @property
    def x509_valid(self) -> bool:
        return self.__x509_valid

    @x509_valid.setter
    def x509_valid(self, valid:bool):
        self.__x509_valid = valid

    @property
    def web_valid(self) -> bool:
        return self.__web_valid

    @web_valid.setter
    def web_valid(self, valid:bool):
        self.__web_valid = valid

    @property
    def sync(self) -> bool :
        return self.__sync_active

    @sync.setter
    def sync(self, activated:bool) -> bool:
        self.__sync_active = activated

    def compute_effective_state(self):
        patched_config = copy.deepcopy(self.transport_config)

        if not self.__sync_active or not self.__x509_valid:
            patched_config.peer_to_peer.ble.enabled = False
            patched_config.peer_to_peer.lan.enabled = False
            patched_config.listen.tcp.enabled = False
            patched_config.connect.tcp_servers = set()

        if not self.__sync_active or not self.__web_valid:
            patched_config.listen.http.enabled = False

        if (self.__identity.is_cloud_sync_enabled() and
                self.__sync_active):
            cloud_url = self.__identity._sync_url()
            patched_config.connect.websocket_urls.add(cloud_url)

        if (patched_config.peer_to_peer.lan.enabled and
                not patched_config.listen.tcp.enabled):
            patched_config.listen.tcp.enabled = True
            patched_config.listen.tcp.interface_ip = "[::]"
            patched_config.listen.tcp.port = 0

        return SyncState(patched_config, self.__identity)



from dittolive import Ditto, OfflinePlayground

from utils import TEST_LICENSE, create_offline_ditto

def test_restart_ditto():
    ditto = create_offline_ditto(working_dir = "/tmp/dittorestartsync")
    ditto.start_sync()
    ditto.stop_sync()
    del ditto
    ditto = create_offline_ditto(working_dir = "/tmp/dittorestartsync")

    ditto.start_sync()
    ditto.stop_sync()

def test_start_sync():
    ditto = create_offline_ditto(working_dir = "/tmp/dittostartsync")

    ditto.start_sync()
    ditto.start_sync()
    ditto.stop_sync()
    ditto.stop_sync()
    ditto.start_sync()
    ditto.stop_sync()


"""Bunch of utilitary functions for dittolive.

If this file grows too much, feel free to split it in a new module.
"""
import weakref

from _dittoffi import ffi, lib
from cbor2 import encoder

global_weakkeydict = weakref.WeakKeyDictionary()

def char_p(string: str) -> bytes:
    """FFI utils function.

    Args:
        string: the string to convert to bytes
    Returns: a char* representation of a Python String with UTF-8 encoding.
    """
    bytes = string.encode("utf-8")
    if 0 in bytes[:-1]:
        raise ValueError("Inner null byte in string")
    return bytes

def empty_char_p():
    return ffi.cast("char*", ffi.NULL)

def string(charp) -> str:
    """Turn char pointer into regular Python string.

    Args:
        charp (_type_): pointer to C utf-8 string

    Returns:
        str: Python valid string
    """
    return ffi.string(charp).decode()

def bytes_from_boxed(slice_boxed_uint8_t) -> bytes:
    """Create a Python `bytes` from a slice_boxed_uint8_t.

    Args:
        slice_boxed_uint8_t : boxed slice return by the FFI

    Returns:
        bytes: Python byte repr
    """
    length = slice_boxed_uint8_t.len
    ptr = slice_boxed_uint8_t.ptr
    # CHECKME(Ronan) this seems a questionnable solution, almost
    # looking like an HACK

    if ptr == ffi.NULL:
        return None

    slice = bytes([ptr[i] for i in range(0,length)])
    lib.ditto_c_bytes_free(slice_boxed_uint8_t)
    return slice

def slice_ref_uint8(data:bytes):
    # CHECKME : use it with `with slice(data) as ...`
    # to avoid cbor getting gargabe collected early
    """Return a FFI slice.

    You need to add [0] to access the value.
    """
    slice_ref = ffi.new("slice_ref_uint8_t*")
    length = len(data)
    data = ffi.new("uint8_t[]", data)
    slice_ref.ptr = data
    slice_ref.len = length
    return slice_ref

def cbor_slice_ref_uint8(data:object):
    """Return a non empty FFI slice representing CBOR data.

    You need to add [0] to access the value.
    """
    # CHECKME : use it with `with slice(data) as ...`
    # to avoid cbor getting gargabe collected early
    """Return a FFI slice."""
    cbor = encoder.dumps(data)
    return slice_ref_uint8(cbor)

def empty_slice_ref_uint8(pointer = ffi.NULL):
    """Return a FFI slice."""
    slice_ref = ffi.new("slice_ref_uint8_t*", dict(
        len = 0,
        ptr = pointer
    ))
    return slice_ref

def empty_cbor_slice_ref_uint8():
    """Return a FFI slice.

    You need to add [0] to access the value.
    """
    return empty_slice_ref_uint8(ffi.cast("uint8_t*", 0xbad000))

def empty_order_slice_ref():
    """Return a FFI slice.

    You need to add [0] to access the value.
    """
    slice_ref = ffi.new("slice_ref_COrderByParam_t*", dict(
        len = 0,
        ptr = ffi.cast("COrderByParam_t*" ,0xbad000)
    ))
    return slice_ref



"""Document Id class."""


from _dittoffi import lib
from dittolive._utils import slice_ref_uint8, string

class DocumentId():
    """Document id class."""
    def __init__(self, id_bytes) -> None:
        """Creates a new DocumentId with the given bytes value."""
        self.__bytes = id_bytes

    def _raw(self):
        """Returns a raw representation of the DocumentId for the FFI."""
        return slice_ref_uint8(self.__bytes)

    def __str__(self) -> str:
        """Return a query compatible string view of the DocumentId."""
        doc_id_slice = self._raw()
        raw_bytes = lib.ditto_document_id_query_compatible(doc_id_slice[0],
                                    lib.STRING_PRIMITIVE_FORMAT_WITHOUT_QUOTES)

        return string(raw_bytes)

    def __eq__(self, other: object) -> bool:
        """Return True is two DocumentId are equal.

        Args:
            other (object): Any object

        Returns:
            bool: True if the DocumentId are equal.
        """
        if not isinstance(other, DocumentId):
            return False
        return self.__bytes == other.__bytes



"""Ditto module."""

import platform
import os
from _dittoffi import lib

from .identity import Identity
from .log import LogLevel
from .store import Store
from ._utils import char_p
from .exceptions import NotActivatedException, ConfigException, LicenseException
from ._transports import TransportConfig, TransportSync
from contextlib import contextmanager

def _sanitize_working_dir(working_dir:os.path):
    """Compute the effective Ditto working directory."""
    platform_name = platform.system()
    if not working_dir:
        if platform_name in ("Linux", "Darwin"):
            home = os.environ.get("HOME")
            working_dir = home + "/Documents/"
        elif platform_name == "Windows":
            raise NotImplementedError
        else:
            # Not supported platform
            raise NotImplementedError
    elif platform_name in ("Linux", "Darwin"):
        working_dir += "/"
    elif platform_name == "Windows":
        working_dir += "\\"
    else:
        # Not supported platform
        raise NotImplementedError
    working_dir += "ditto"
    return working_dir

class Ditto:
    """Ditto entry point."""
    def __init__(self, identity:Identity, *, working_dir:str = "",
            transport_config: TransportConfig = None):
        """Create a new Ditto instance.

        Args:
            identity (Identity): Identity used to identify the Ditto peer
            working_dir (str, optional): Directory where Ditto will persist data.
                Defaults to "".
            transport_config (TransportConfig, optional): _description_.
                Defaults to TransportConfig().
        """
        # Get local folder for Linux
        working_dir = _sanitize_working_dir(working_dir)

        # create uninitialized ditto
        uninitialized_ditto = lib.uninitialized_ditto_make(char_p(working_dir))
        # get auth client from identity
        identity._authenticate(working_dir)
        auth_client = identity.auth_client()
        # create real raw handle to ditto
        raw_ditto = lib.ditto_make(uninitialized_ditto, auth_client,
            lib.HISTORY_TRACKING_DISABLED)

        self.__activated = not identity.requires_offline_license_token()
        self.__store = Store(raw_ditto)
        if transport_config is None:
            transport_config = TransportConfig()

        if identity._is_online():
            transport_config.connect.websocket_urls.add(identity._sync_url())

        self.__transports = TransportSync(transport_config, raw_ditto, identity)
        self.__identity = identity
        self.__raw_ditto = raw_ditto

    def __del__(self):
        """Delete the ditto instance."""
        self.close()
        lib.ditto_free(self.__raw_ditto)

    def close(self):
        """Close the current ditto instance."""
        self.__identity.close()
        lib.ditto_stop_tcp_server(self.__raw_ditto)
        lib.ditto_shutdown(self.__raw_ditto)

    def set_license_from_env(self, var_name:str):
        """Load the license from a environment variable."""
        license_token = os.environ.get(var_name)

        if not license_token:
            raise ConfigException

        self.set_offline_only_license_token(license_token)

    def set_offline_only_license_token(self, license_token:str):
        """Set offline license token.

        Args:
            license_token (str): Content of the license

        Raises:
            LicenseException: Explain why license has been refused
        """
        # We are not interested in error messages for now

        err_msg = []
        c_license_token = char_p(license_token)

        result = lib.verify_license(c_license_token, err_msg )

        if result == lib.LICENSE_VERIFICATION_RESULT_LICENSE_OK:
            self.__activated = True
            return

        raise LicenseException

    def start_sync(self):
        """Start sync with other peers."""
        if not self.__activated and self.__identity.requires_offline_license_token():
            raise NotActivatedException
        self.__transports.start_sync()

    def stop_sync(self):
        """Stop sync with other peers."""
        self.__transports.stop_sync()

    def is_sync_active(self) -> bool:
        """Returns a flag indicating whether sync is active.

        Use `start_sync` to activate sync and `stop_sync` to deactivate sync.
        """
        raise NotImplementedError

    def authenticator(self):
        """Provides access to Authenticator.

        Use it for information and methods for logging on to Ditto Cloud.
        """

    def presence(self):
        """Provides access to the SDK's presence functionality."""
        raise NotImplementedError

    def disk_usage(self):
        """Provides access to the SDK's disk usage."""
        raise NotImplementedError

    @property
    def transport_config(self) -> TransportConfig:
        """Returns the current transport config."""
        return self.__transports.transport_config

    @contextmanager
    def open_transport_config(self):
        """Assign a new transports configuration.

        By default peer-to-peer transports (Bluetooth and WiFi, where available)
        are enabled.

        You may use this method to alter the configuration at any time. Sync will
        not begin until @ref start_sync is called.
        """
        try:
            yield self.transport_config
        finally:
            self.__transports._update()

    @property
    def store(self):
        """Give access to the Store."""
        return self.__store

    def is_logging_enabled(self) -> bool:
        """Gets whether or not logging is enabled for Ditto."""
        raise NotImplementedError

    def set_logging_enabled(self, enabled: bool):
        """Sets whether or not logging is enabled for Ditto."""
        raise NotImplementedError

    def emoji_log_level_headings_enabled(self) -> bool:
        """Gets whether emoji log level headings are enabled for Ditto."""
        raise NotImplementedError

    def set_emoji_log_level_heading_enabled(self, enabled: bool):
        """Sets whether emoji log level headings are enabled for Ditto."""
        raise NotImplementedError

    def minimum_log_level(self) -> LogLevel:
        """Gets the minimum log level at which logs will be emitted for Ditto."""
        raise NotImplementedError

    def set_minimum_log_level(self, log_level: LogLevel):
        """Sets the minimum log level at which logs should be emitted for Ditto."""
        raise NotImplementedError

    def app_id(self) -> str:
        """Gets the Ditto application Id."""
        raise NotImplementedError

    def persistence_directory(self) -> str:
        """Gets the persistence directory used by Ditto to persist data."""
        raise NotImplementedError



from utils import create_offline_ditto

def test_get_collection():
    ditto = create_offline_ditto()
    store = ditto.store
    store.collection("collection_name")
    store["collection_name"]

def test_collections():
    ditto = create_offline_ditto()
    store = ditto.store
    store.collection("collection_name1")
    store.collection("collection_name2")
    store.collection("collection_name3")


"""Result returned when updating a document."""

from enum import Enum
from .document_id import DocumentId

class UpdateResultType(Enum):
    """Type of the update."""
    Set = 0,
    Removed = 1,

class UpdateResult:
    """Generic update result."""
    def __init__(self, doc_id: DocumentId, path:str, operation_type: UpdateResultType):
        """Create a new UpdateResult."""
        self.doc_id = doc_id
        self.path = path
        self.operation_type = operation_type

class UpdateResultSet(UpdateResult):
    """Result returned when setting a value in a mutable document."""
    def __init__(self, doc_id: DocumentId, path: str, value):
        """Create a new UpdateResultSet."""
        super().__init__(doc_id, path, UpdateResultType.Set)
        self.value = value

class UpdateResultRemoved(UpdateResult):
    """Result returned when removing a value from a mutable document."""
    def __init__(self, doc_id: DocumentId, path: str):
        """Create a new UpdateResultRemoved."""
        super().__init__(doc_id, path, UpdateResultType.Removed)



from dittolive import Ditto
from dittolive import TransportConfig

from utils import create_offline_ditto, create_online_ditto

def test_enable_ble():
    transport_config = TransportConfig()
    transport_config.ble_enabled = True
    ditto = create_offline_ditto(transport_config=transport_config)
    ditto.start_sync()
    with ditto.open_transport_config() as config:
        config.bluetooth_enabled = False

def test_edit_lan():
    transport_config = TransportConfig()
    transport_config.lan_enabled = True
    ditto = create_offline_ditto(transport_config=transport_config)
    ditto.start_sync()
    with ditto.open_transport_config() as config:
        config.lan_enabled = False

def test_edit_tcp():
    transport_config = TransportConfig()
    transport_config.tcp_enabled = True
    transport_config.listen.tcp.interface_ip = "127.0.0.1"
    transport_config.listen.tcp.port = "16042"
    transport_config.connect.tcp_servers.add("127.0.0.1:16041")
    ditto = create_offline_ditto(transport_config=transport_config)
    ditto.start_sync()
    with ditto.open_transport_config() as config:
        config.tcp_enabled = False

def test_edit_ws():
    transport_config = TransportConfig()
    ditto = create_online_ditto(transport_config=transport_config)
    ditto.start_sync()
    with ditto.open_transport_config() as config:
        config.connect.websocket_urls.add("wss://brand_new_websocket.ditto.live")

    with ditto.open_transport_config() as config:
        config.connect.websocket_urls.remove("wss://brand_new_websocket.ditto.live")

    ditto.stop_sync()



"""Store module."""

from _dittoffi import lib

from dittolive._utils import char_p
from .collection import Collection

class Store:
    """Database for ditto."""

    def __init__(self, ditto):
        """Initialize store."""
        self.__raw_ditto = ditto

    def collection(self, collection_name: str) -> Collection:
        """Return collection named 'collection_name'."""
        byted_collection_name = char_p(collection_name)

        status = lib.ditto_collection(self.__raw_ditto, byted_collection_name)

        if status != 0:
            raise KeyError(collection_name)

        return Collection(self.__raw_ditto, byted_collection_name)

    def __getitem__(self, collection_name: str) -> Collection:
        """Return collection named 'collection_name'."""
        return self.collection(collection_name)

    def write(self, query: str):
        """Write transaction."""
        raise NotImplementedError

    def collections(self):
        """Returns  a PendingCollectionsOperation."""
        raise NotImplementedError



"""Subscription to Ditto Documents."""

from dittolive.__observer import Observer
from _dittoffi import lib
from dittolive._utils import empty_order_slice_ref, empty_slice_ref_uint8
from dittolive.exceptions import FfiException
class Subscription(Observer):
    """Subscription class.

    Subscriptions are used to register interest in receiving updates
    for specified documents.
    """
    def __init__(self, ditto_raw, collection_name:bytes, query:bytes, *,
                 query_args=None, order_by=None, limit:int=-1, offset:int=0):
        """Create a new Subscription.

        Keep this alive for as long as you want
        to receive updates on the specified documents.

        Args:
            ditto_raw (_type_): raw pointer to Ditto
            collection_name (bytes): char_p pointer to the collection name
            query (bytes): char_p to the query string
            query_args (_type_): _description_
            order_by (_type_): _description_
            limit (int, optional): Maximum number of documents. Defaults to -1.
            offset (int, optional): Offset of the document to retrieve. Defaults to 0.
        """
        super().__init__()
        self.__raw_ditto = ditto_raw
        self.__collection_name = collection_name
        self.__query = query

        # TODO: handle query_args and order_by when we add the new `Collection::find`
        # API for DQL
        query_args = empty_slice_ref_uint8()
        order_by = empty_order_slice_ref()

        self.__query_args = query_args
        self.__order_by = order_by
        self.__limit = limit
        self.__offset = offset

        res = lib.ditto_add_subscription(self.__raw_ditto, self.__collection_name,
                        self.__query, self.__query_args[0],
                        self.__order_by[0], self.__limit, self.__offset)

        if res != 0:
            raise FfiException

    def _close_callback(self):
        res = lib.ditto_remove_subscription(self.__raw_ditto, self.__collection_name,
                        self.__query, self.__query_args[0],
                        self.__order_by[0], self.__limit, self.__offset)
        if res != 0:
            raise FfiException



from abc import ABC, abstractmethod

class Observer(ABC):
    def __init__(self):
        self.__observer_closed = False

    def __del__(self):
        self.close()

    @abstractmethod
    def _close_callback(self):
        """Stop the observer."""

    def close(self):
        """Stop the observer."""
        if not self.__observer_closed:
            self._close_callback()
            self.__observer_closed = True



"""Identity are used to identify the peers in a Ditto mesh network."""
from .identity import Identity
from .offline_playground import OfflinePlayground
from .online_playground import OnlinePlayground

__all__ = ["Identity", "OfflinePlayground", "OnlinePlayground"]



"""OnlinePlayground for testing sync."""

from _dittoffi import lib

from .identity import Identity
from dittolive._utils import char_p
from dittolive.exceptions import FfiException

class OnlinePlayground(Identity):
    """Class OnlinePlayground."""
    def __init__(self, app_id: str, shared_token:str, *, custom_auth_url:str=None,
                cloud_sync_enabled:bool = False):
        """Create a new OnlinePlayground.

        Args:
            app_id (str): application id, as an UUID string
            shared_token (str): Token shared between apps
            custom_auth_url (str, optional): Custom URL for authentication.
                Defaults to None.
            cloud_sync_enabled (bool, optional): Should this peer sync with Ditto cloud.
                Default to False
        """
        super().__init__(app_id=app_id)
        self.__shared_token = shared_token
        self.__custom_auth_url = custom_auth_url
        self.__cloud_sync_enabled = cloud_sync_enabled

    def _authenticate(self, working_dir: str):
        working_dir_bytes = char_p(working_dir)
        app_id_bytes = char_p(self.app_id())
        shared_token_bytes = char_p(self.__shared_token)

        if self.__custom_auth_url:
            url_bytes = char_p(self.__custom_auth_url)
        else:
            url_bytes = char_p(self._auth_url())

        result = lib.ditto_auth_client_make_anonymous_client(
            working_dir_bytes,
            app_id_bytes,
            shared_token_bytes,
            url_bytes)

        if result.status_code != 0:
            raise FfiException

        self._auth_client = result.auth_client

    def is_cloud_sync_enabled(self):
        """Return true is the connection with Ditto cloud is enabled."""
        return self.__cloud_sync_enabled

    def _is_online(self):
        """Return true is the connection with Ditto cloud is enabled."""
        return True



from dittolive.__observer import Observer

class FooObserver(Observer):
    def __init__(self, closed):
        super().__init__()
        self.closed = closed

    def _close_callback(self):
        self.closed[0] = True
        ...

def test_del_observer():
    closed = [False]
    observer = FooObserver(closed)
    assert closed[0] == False
    del observer
    assert closed[0] == True

def test_close_observer():
    closed = [False]
    observer = FooObserver(closed)
    assert closed[0] == False
    observer.close()
    assert closed[0] == True

def test_close_del_observer():
    closed = [False]
    observer = FooObserver(closed)
    assert closed[0] == False
    observer.close()
    assert closed[0] == True
    del observer
    assert closed[0] == True

def test_scoped_observer():
    def observe(closed):
        assert closed[0] == False
        _ = FooObserver(closed)

    closed = [False]
    observe(closed)
    assert closed[0] == True


"""Document module."""
import cbor2

from .document_path import DocumentPath
from .document_id import DocumentId
from _dittoffi import lib
from dittolive._utils import bytes_from_boxed
class Document:
    """Document."""
    def __init__(self, doc_ptr):
        """Create a new Document."""
        self.__doc_ptr = doc_ptr

    def __del__(self):
        """Delete the document and release ressources."""
        lib.ditto_document_free(self.__doc_ptr)

    def id(self) -> DocumentId:
        """Get the document's Id."""
        boxed_slice = lib.ditto_document_id(self.__doc_ptr)
        doc_id = bytes_from_boxed(boxed_slice)
        return DocumentId(doc_id)

    @property
    def value(self) -> dict:
        """Get the document's inner value."""
        cbor = lib.ditto_document_cbor(self.__doc_ptr)
        cbor_bytes = bytes_from_boxed(cbor)

        return cbor2.loads(cbor_bytes)

    def get(self, key:str) -> DocumentPath:
        """Follow path in a document.

        Use to specify a path to a key in the document that you can
        subscript further to access a nested key in the document.
        """
        return DocumentPath(self.__doc_ptr, self.id(), key)

    def __getitem__(self, key:str) -> DocumentPath:
        """Follow path in a document.

        Use to specify a path to a key in the document that you can
        subscript further to access a nested key in the document.
        """
        return self.get(key)



"""LiveQuery are used to monitore documents changes."""

from typing import Callable, List

from dittolive.__observer import Observer
from dittolive._utils import empty_slice_ref_uint8, empty_order_slice_ref
from dittolive.exceptions import FfiException
from _dittoffi import ffi, lib
from .document import Document

UNASSIGNED_ID_SENTINEL: int = -1

class LiveQueryMove():
    """Live Query Move.

    Describes the index in a list of documents that a document was previously found at
    `origin` and the index that it can now be found at `to`.
    """
    def __init__(self, origin, to):
        """Creates a new LiveQuery move."""
        self._left = origin
        self._right = to

class _LiveQueryEvent():
    def __init__(self,*, is_initial = False) -> None:
        self._is_initial = is_initial

    def is_initial(self) -> bool:
        return self._is_initial
class _LiveQueryInitial(_LiveQueryEvent):
    def __init__(self) -> None:
        super().__init__(is_initial = True)

class _LiveQueryUpdate(_LiveQueryEvent):
    def __init__(self, c_cb_params):
        super().__init__(is_initial = False)
        #TODO : Params here are not ready to be used

        raw_vec_documents = c_cb_params.old_documents
        docs_ptr = raw_vec_documents.ptr
        docs_number = raw_vec_documents.len
        self.old_documents = [Document(docs_ptr[i]) for i in range(docs_number)]

        #TODO
        # self.insertions = c_cb_params.insertions
        # self.deletions = c_cb_params.deletions
        # self.updates = c_cb_params.updates
        # self.moves = [LiveQueryMove(left, right) for (left,right)
        #   in groupby(c_cb_params.moves, 2)] # TODO

# TODO: This should be an event callback list
# WITH the list of Documents inside
LiveQueryDocumentCallback = Callable[[List[Document], _LiveQueryEvent], None]

@ffi.callback("void(void*, c_cb_params_t)")
def c_live_query_event_handler(live_query_ptr, params):
    """C callback for live query. This is passed to the Rust core through the FFi."""
    live_query = ffi.from_handle(live_query_ptr)
    #compute documents
    raw_vec_documents = params.documents
    docs_ptr = raw_vec_documents.ptr
    docs_number = raw_vec_documents.len
    documents = [Document(docs_ptr[i]) for i in range(docs_number)]

    # handle event
    if params.is_initial:
        live_query.callback(documents, _LiveQueryInitial())
    else:
        event = _LiveQueryUpdate(params)
        live_query.callback(documents, event)
        lib.ditto_only_vec_documents_free(params.old_documents)

    lib.ditto_free_indices(params.insertions)
    lib.ditto_free_indices(params.deletions)
    lib.ditto_free_indices(params.updates)
    lib.ditto_free_indices(params.moves)
    lib.ditto_only_vec_documents_free(params.documents)

    #TODO Free fields in params

class LiveQuery(Observer):
    """LiveQuery."""
    def __init__(self, raw_ditto, query: bytes, coll_name: bytes,
                 callback: LiveQueryDocumentCallback,
                 *, query_args=None, order_by=None,
                 limit: int = -1, offset: int = 0,
                 availability=lib.LIVE_QUERY_AVAILABILITY_ALWAYS):
        """Creates a new LiveQuery.

        Args:
            raw_ditto (_type_): _description_
            query (bytes): _description_
            coll_name (bytes): _description_
            callback (LiveQueryDocumentCallback): _description_
            query_args (_type_, optional): _description_. Defaults to None.
            order_by (_type_, optional): _description_. Defaults to None.
            limit (int, optional): _description_. Defaults to -1.
            offset (int, optional): _description_. Defaults to 0.
            availability (_type_, optional): _description_.
                Defaults to lib.LIVE_QUERY_AVAILABILITY_ALWAYS.

        Raises:
            FfiException: _description_
        """
        super().__init__()
        handle = ffi.new_handle(self)

        self.__raw_ditto = raw_ditto
        self.__query = query
        self.__collection_name = coll_name,
        self.callback = callback
        self.__handle =handle

        args = empty_slice_ref_uint8()
        order_by = empty_order_slice_ref()

        res = lib.ditto_live_query_register_str(
            raw_ditto,
            coll_name,
            query,
            args[0],  # args params
            order_by[0],  # Order by
            limit,
            offset,
            availability,
            handle,  # ptr to self
            # TODO: add proper retain/release mechanism
            ffi.NULL,  # retain
            ffi.NULL,  # release
            c_live_query_event_handler  # c callback
        )

        if res.status_code != 0:
            raise FfiException

        self.__id = res.i64

        if self.__id == UNASSIGNED_ID_SENTINEL:
            raise ValueError("Not a valid LiveQuery ID")
            # LiveQuery was not given a valid id
        else:
            pass
            # TODO : log this
            # Log valid query id

        lib.ditto_live_query_start(raw_ditto, self.__id)

    def _close_callback(self):
        """Drop the live query."""
        # TODO : log this
        lib.ditto_live_query_stop(self.__raw_ditto, self.__id)



import pytest
import time

from dittolive import *
from utils import create_offline_ditto
from dittolive.exceptions import DocumentNotFoundException
from asyncio.futures import Future
import asyncio

def test_subscribe():
    ditto = create_offline_ditto()
    collection = ditto.store["test"]
    doc_id = collection.upsert({"content": "value"})
    sub = collection.find_by_id(doc_id).subscribe()
    assert sub
    sub.close()

@pytest.mark.asyncio
async def test_observe_lambda():
    ditto = create_offline_ditto()
    collection = ditto.store["test"]
    doc_id = collection.upsert({"content": "value"})
    future = Future()

    callback = lambda doc, event : future.set_result(True)
    live_query = collection.find_by_id(doc_id).observe(callback)
    doc_id = collection.upsert({"_id": str(doc_id), "content": "New value"})
    await future
    live_query.close()

@pytest.mark.asyncio
async def test_observe_scoped_function_by_upsert():
    future = Future()
    counter = 0
    def scoped_function(doc, event):
        nonlocal counter
        if counter == 0:
            counter += 1
        else:
            future.set_result(True)

    ditto = create_offline_ditto()
    collection = ditto.store["test"]
    doc_id = collection.upsert({"content": "value"})

    live_query = collection.find_by_id(doc_id).observe(scoped_function)

    doc_id = collection.upsert({"_id": str(doc_id), "content": "New value"})
    while not future.done(): # Directly calling `await future` locks the thread
        await asyncio.sleep(0.01)
    live_query.close()

@pytest.mark.asyncio
async def test_observe_scoped_function_by_update():
    future = Future()
    counter = 0

    def scoped_function(doc, event):
        nonlocal counter
        if counter == 0:
            counter += 1
        else:
            future.set_result(True)

    ditto = create_offline_ditto()
    collection = ditto.store["test"]
    doc_id = collection.upsert({"content": "value"})

    live_query = collection.find_by_id(doc_id).observe(scoped_function)

    updater = lambda mut_doc: mut_doc["new_content"].set("BrandNewSDK")
    collection.find_by_id(doc_id).update(updater)

    while not future.done(): # Directly calling `await future` locks the thread
        await asyncio.sleep(0.01)
    live_query.close()

def test_remove():
    ditto = create_offline_ditto()
    collection = ditto.store["test"]

    doc_id = collection.upsert({"content": "value"})
    doc = collection.find_by_id(doc_id).exec()

    collection.find_by_id(doc_id).remove()

    with pytest.raises(DocumentNotFoundException):
        doc = collection.find_by_id(doc_id).exec()

    collection.find_by_id(doc_id).remove()

def test_evict():
    ditto = create_offline_ditto()
    collection = ditto.store["test"]

    doc_id = collection.upsert({"content": "value"})
    doc = collection.find_by_id(doc_id).exec()

    collection.find_by_id(doc_id).evict()
    with pytest.raises(DocumentNotFoundException):
        doc = collection.find_by_id(doc_id).exec()
    collection.find_by_id(doc_id).evict()

def test_exec():
    ditto = create_offline_ditto()
    collection = ditto.store["test"]

    doc_id = collection.upsert({"content": "value"})
    doc = collection.find_by_id(doc_id).exec()
    doc != None

def test_update_lambda():
    ditto = create_offline_ditto()
    collection = ditto.store["test"]
    doc_id = collection.upsert({"content": "value"})
    doc = collection.find_by_id(doc_id).exec()
    assert doc["content"].value == "value"
    cb = lambda mut_doc : mut_doc["content"].set(3)
    result = collection.find_by_id(doc_id).update(cb)

    #TODO result
    doc = collection.find_by_id(doc_id).exec()
    assert doc["content"].value == 3

def test_update_function():
    ditto = create_offline_ditto()
    collection = ditto.store["test"]

    def updater(mut_doc):
        mut_doc["content"] = 3

    doc_id = collection.upsert({"content": "value"})
    doc = collection.find_by_id(doc_id).exec()
    assert doc["content"].value == "value"
    result = collection.find_by_id(doc_id).update(updater)

    #TODO result
    doc = collection.find_by_id(doc_id).exec()
    assert doc["content"].value == 3


"""PendingIdSpecificOperation."""

from typing import List, Callable

from _dittoffi import lib

from .document import Document
from .__transaction import WriteTransaction
from .document_id import DocumentId
from .mutable_document import MutableDocument
from dittolive.exceptions import FfiException, DocumentNotFoundException,\
    NOT_FOUND_ERROR_CODE
from dittolive._utils import char_p
from .live_query import LiveQuery
from .subscription import Subscription
from .update_result import UpdateResult


class _SingleDocumentLiveQueryEvent:
    def __init__(self, is_initial:bool, old_document:Document = None):
        self.is_initial = is_initial
        self.old_document = old_document

LiveQuerySingleDocumentCallback = Callable[[MutableDocument,
                                            _SingleDocumentLiveQueryEvent], None]

class PendingIdSpecificOperation:
    """_summary_."""

    def __init__(self, raw_ditto, collection_name:bytes, doc_id:DocumentId):
        """Create a new PendingIdSpecificOperation.

        Args:
            raw_ditto (_type_): handle to CDitto
            collection_name (bytes): Collection name in char_p ~ bytes
            doc_id (DocumentId): DocId
        """
        self.__raw_ditto = raw_ditto
        self.__collection_name = collection_name
        self.__doc_id = doc_id

    def __query(self) -> bytes:
        """Return a char_p repr of the inner query.

        Returns:
            bytes: char_p of the specific id query
        """
        query = f" _id == {str(self.__doc_id)!r}"
        return char_p(query)

    def subscribe(self) -> Subscription:
        """Create new subscription.

        Returns:
            Subscription: _description_
        """
        return Subscription(self.__raw_ditto, self.__collection_name, self.__query())

    def remove(self):
        """Remove selected document.

        Raises:
            FfiException: The document could not be removed
        """
        with WriteTransaction(self.__raw_ditto) as write_txn:
            was_removed = lib.ditto_collection_remove(
                self.__raw_ditto,
                self.__collection_name,
                write_txn,
                self.__doc_id._raw()[0],
            )
            if not was_removed:
                lib.ditto_write_transaction_rollback(self.__raw_ditto, write_txn)
                raise FfiException

    def evict(self):
        """Evict selected document.

        Raises:
            NotImplementedError: _description_
        """
        with WriteTransaction(self.__raw_ditto) as write_txn:
            was_removed = lib.ditto_collection_evict(
                self.__raw_ditto,
                self.__collection_name,
                write_txn,
                self.__doc_id._raw()[0],
            )

            if not was_removed:
                lib.ditto_write_transaction_rollback(self.__raw_ditto, write_txn)
                raise FfiException

    def exec(self) -> Document:
        """Get selected document.

        Raises:
            FfiException: _description_
            DocumentNotFoundException: _description_

        Returns:
            Document: _description_
        """
        doc_ptr = self._exec()
        return Document(doc_ptr)

    def _exec(self):
        """Internal exec.

        Return document pointer
        """
        read_txn = lib.ditto_read_transaction(self.__raw_ditto).txn

        result = lib.ditto_collection_get(
            self.__raw_ditto,
            self.__collection_name,
            self.__doc_id._raw()[0],
            read_txn
        )

        # TODO : not found error code
        if result.status_code == NOT_FOUND_ERROR_CODE:
            raise DocumentNotFoundException
        elif result.status_code != 0:
            raise FfiException

        return result.document

    def observe(self, callback: LiveQuerySingleDocumentCallback) -> LiveQuery:
        """Observe selected document.

        Raises:
            NotImplementedError: _description_

        Returns:
            LiveQuery: _description_
        """

        def lq_doc_event_handler(documents, event):
            """Callback for LiveQuery. Wrap the user callback.

            Args:
                documents (_type_): List of raw Documents
                event (_type_): LiveQueryEvent
                signal_next (_type_): Next signal
            """
            old_document = None
            new_document = None

            if not event.is_initial and len(event.old_documents) == 1:
                old_document = event.old_documents[0]

            if len(documents) == 0:
                ...
            elif len(documents) == 1:
                new_document = documents[0]
            else:
                return

            event = _SingleDocumentLiveQueryEvent(old_document=old_document,
                                                is_initial=event.is_initial)

            callback(new_document, event)

        return LiveQuery(self.__raw_ditto, self.__query(),
            self.__collection_name, lq_doc_event_handler)

    def update(self, updater: LiveQuerySingleDocumentCallback) -> List[UpdateResult]:
        """Update selected document with given callback.

        Raises:
            NotImplementedError: _description_

        Returns:
            List[UpdateResult]: _description_
        """
        doc_ptr = self._exec()
        doc = MutableDocument(doc_ptr, self.__doc_id)

        with WriteTransaction(self.__raw_ditto) as write_txn:
            updater(doc)

            lib.ditto_collection_update(self.__raw_ditto, self.__collection_name,
                                        write_txn, doc_ptr)

        return doc.results
